techniques.html

#### Techniques

The preprocessor metaprogramming techniques are presented in example format.

#### Example - Use a local macro to avoid small scale repetition.

    #define BOOST_PP_DEF(op) /* ..................................... */ \
       template<class T, int n>                                          \
       vec<T, n> operator op ## =(vec<T, n> lhs, const vec<T, n>& rhs) { \
          for (int i = 0; i < n; ++i) {                                  \
             lhs(i) op ## = rhs(i);                                      \
          }                                                              \
       }                                                                 \
       /**/

    BOOST_PP_DEF(+)
    BOOST_PP_DEF(-)
    BOOST_PP_DEF(*)
    BOOST_PP_DEF(/)

    #undef BOOST_PP_DEF

Tip: It is usually okay to use a standard macro name like `BOOST_PP_DEF` for this kind of code because the macro is both defined and undefined in the immediate site of its use.
Tip: It is easier to verify proper use of the line continuation operator when they are aligned.
Notes: You can extend this example by defining more and different kinds of operators. Before doing so, consider using the algebraic categories technique introduced in [Barton] or a layered architecture (see for instance [Czarnecki]). However, at some point you must type the operator tokens `*`, `/`, `+`, `-`, etc., because it is impossible to generate them using templates. The resulting categorical repetition of tokens can be eliminated by using preprocessor metaprogramming.

#### Example - Use BOOST_PP_EMPTY as an unused parameter in local macro instantiations.

    #define BOOST_PP_DEF(cv) /* ... */ \
       template<class base>            \
       cv() typename implement_subscript_using_begin_subscript<base>::value_type& \
       implement_subscript_using_begin_subscript<base>::operator[](index_type i) cv() { \
          return base::begin()[i];     \
       }                               \
       /**/

       BOOST_PP_DEF(BOOST_PP_EMPTY)
       BOOST_PP_DEF(BOOST_PP_IDENTITY(const))

How: BOOST_PP_EMPTY() expands to nothing and can be used as an unused parameter.
Note: BOOST_PP_EMPTY with the () never gets expanded. The () is necessary to invoke a function-like macro.
Caveat: You cannot safely use concatenation while using BOOST_PP_EMPTY().
Tip: Occasionally, one or two lines are considerably longer than the rest. It can often save some work to not align all the line continuation operators without making the code too unreadable.
Tip: Use syntax highlighting on preprocessor metaprogramming macro identifiers such as:

- BOOST_PP_DEF
- BOOST_PP_EMPTY
- BOOST_PP_REPEAT
- ...

It can greatly improve readability.

#### Example - Use BOOST_PP_CAT instead of ## when necessary.

    #define STATIC_ASSERT(expr) \
       enum { BOOST_PP_CAT(static_check_, __LINE__) = (expr) ? 1 : -1 }; \
       typedef char \
          BOOST_PP_CAT(static_assert_, __LINE__)[BOOST_PP_CAT(static_check_, __LINE__)] \
       /**/

    // ...

    STATIC_ASSERT(sizeof(int) <= sizeof(long));

Why: Macro expansion proceeds recursively in "layers." Token pasting prevents the preprocessor from performing macro expansion, therefore it is often necessary to delay token concatenation.

#### Example - Use BOOST_PP_STRINGIZE instead of # whenever necessary.

    #define NOTE(str) \
       message(__FILE__ "(" BOOST_PP_STRINGIZE(__LINE__) ") : " str) \
       /**/

    // ...

    #pragma NOTE("TBD!")

Why: Macro expansion proceeds recursively in "layers." Stringization prevents the preprocessor from performing macro expansion, therefore it is often necessary to delay stringization.

#### Example - Use BOOST_PP_ENUM_PARAMS (and its variants) or BOOST_PP_REPEAT and BOOST_PP_COMMA_IF to avoid O(n) repetition on lists in general.

    struct make_type_list_end;

    template<
       BOOST_PP_ENUM_PARAMS_WITH_A_DEFAULT(
          MAKE_TYPE_LIST_MAX_LENGTH,
          class T,
          make_type_list_end
       )
    >
    struct make_type_list {
       private:
          enum { end = is_same<T0, make_type_list_end>::value };
       public:
          typedef typename type_if<
             end, type_cons_empty,
             type_cons<
                T0,
                typename type_inner_if<
                   end, type_identity<end>,
                   make_type_list<
                      BOOST_PP_ENUM_SHIFTED_PARAMS(
                         MAKE_TYPE_LIST_MAX_LENGTH,
                         T
                      )
                   >
                >::type
             >
          >::type type;
    };

How: BOOST_PP_REPEAT uses simulated recursion (pseudo code):

    #define BOOST_PP_REPEAT(n, m, p) BOOST_PP_REPEAT ## n(m, p)

    #define BOOST_PP_REPEAT0(m, p)
    #define BOOST_PP_REPEAT1(m, p) m(0, p)
    #define BOOST_PP_REPEAT2(m, p) m(0, p) m(1, p)
    #define BOOST_PP_REPEAT3(m, p) BOOST_PP_REPEAT2(m, p) m(2, p)
    #define BOOST_PP_REPEAT4(m, p) BOOST_PP_REPEAT3(m, p) m(3, p)
    // ...

Note: This is no longer how BOOST_PP_REPEAT is implemented, so the code above is illustrative only!
BOOST_PP_ENUM_PARAMS and its variations use BOOST_PP_REPEAT. BOOST_PP_COMMA_IF(I) expands to a comma if I != 0. BOOST_PP_INC(I) essentially expands to "I+1," and BOOST_PP_DEC(I) essentially expands to "I-1.".

#### Example - Use a conditional macro definition to enable user configuration of code repetition based on need rather than some "reasonable" upper limit.

    #ifndef MAKE_TYPE_LIST_MAX_LENGTH
    #define MAKE_TYPE_LIST_MAX_LENGTH 8
    #endif

Now the user can configure the `make_type_list` primitive without modifying library code.

#### Example - Use BOOST_PP_REPEAT and a token look-up function to eliminate categorical repetition.

    // CAVEAT:  My compiler is not standard on arithmetic types.
    #define ARITHMETIC_TYPE(I) ARITHMETIC_TYPE ## I

    #define ARITHMETIC_TYPE0    bool
    #define ARITHMETIC_TYPE1    char
    #define ARITHMETIC_TYPE2    signed char
    #define ARITHMETIC_TYPE3    unsigned char
    #define ARITHMETIC_TYPE4    short
    #define ARITHMETIC_TYPE5    unsigned short
    #define ARITHMETIC_TYPE6    int
    #define ARITHMETIC_TYPE7    unsigned int
    #define ARITHMETIC_TYPE8    long
    #define ARITHMETIC_TYPE9    unsigned long
    #define ARITHMETIC_TYPE10   float
    #define ARITHMETIC_TYPE11   double
    #define ARITHMETIC_TYPE12   long double

    #define ARITHMETIC_TYPE_CNT 13

    // ...

    #define BOOST_PP_DEF(z, I, _) /* ... */ \
       catch (ARITHMETIC_TYPE(I) t) {       \
          report_typeid(t);                 \
          report_value(t);                  \
       }                                    \
       /**/

    BOOST_PP_REPEAT(ARITHMETIC_TYPE_CNT, BOOST_PP_DEF, _)

    #undef BOOST_PP_DEF

Note: The repetition of the above example can be eliminated using template metaprogramming [Czarnecki] as well. However categorical repetition of operator tokens cannot be completely eliminated by using template metaprogramming.

#### Example - Use BOOST_PP_REPEAT to avoid O(n*n) repetition.

    #ifndef MAX_VEC_ARG_CNT
    #define MAX_VEC_ARG_CNT 8
    #endif

    // ...

    #define ARG_FUN(z, i, _) BOOST_PP_COMMA_IF(i) T a ## i
    #define ASSIGN_FUN(z, i, ) (*this)[i] = a ## i;

    #define DEF_VEC_CTOR_FUN(z, i, _) /* ... */ \
       vec(BOOST_PP_REPEAT(i, ARG_FUN, _)) { \
          BOOST_PP_REPEAT(i, ASSIGN_FUN, _)  \
       }                                     \
       /**/

    BOOST_PP_REPEAT(BOOST_PP_INC(MAX_VEC_ARG_CNT), DEF_VEC_CTOR_FUN, _)

    #undef ARG_FUN
    #undef ASSIGN_FUN
    #undef DEF_VEC_CTOR_FUN

    // ...

How: BOOST_PP_REPEAT is implemented is a special way to enable automatic recursion.

#### Example - Use BOOST_PP_IF to implement special case for the first element.

    #define COMMA_IF(c) \
       BOOST_PP_IF(c, BOOST_PP_COMMA, BOOST_PP_EMPTY)() \
       /**/

    BOOST_PP_IF(0, true, false) == false;
    BOOST_PP_IF(1, true, false) == true;

BOOST_PP_IF enables convenient generation of lists using BOOST_PP_REPEAT.
Note: THEN and ELSE don't have to be macros. However, if at least one of them is a function-like macro, and you want it be expanded conditionally, you have to make the other parameter a function-like macro too. This can often be done using BOOST_PP_IDENTITY. Consider the following example (by Aleksey Gurtovoy):

    #define NUMBERED_EXPRESSION(i, x) /* ... */ \
       BOOST_PP_IF(                             \
          i,                                    \
          BOOST_PP_IDENTITY(x ## i)             \
          BOOST_PP_EMPTY                        \
       )()                                      \
       /**/

Note: Like in the above implementation of COMMA_IF, the result of BOOST_PP_IF is often invoked and not the THEN and ELSE parameters. If the parameters were invoked, the code would not expand correctly, because the BOOST_PP_EMPTY parameter would get expanded to nothing before the BOOST_PP_IF would be properly expanded.
How: BOOST_PP_IF is defined for the entire repeat range (psuedo code):

    #define BOOST_PP_IF(c, THEN, ELSE) BOOST_PP_IF ## c(THEN, ELSE)

    #define BOOST_PP_IF0(THEN, ELSE) ELSE
    #define BOOST_PP_IF1(THEN, ELSE) THEN
    #define BOOST_PP_IF1(THEN, ELSE) THEN
    // ...

#### Example: Use arithmetic, logical, and comparison operations when necessary.

    #define SPECIAL_NUMBERED_LIST(n, i, elem, special) \
       BOOST_PP_ASSERT_MSG(                     \
          BOOST_PP_LESS(i, n),                  \
          bad params for SPECIAL_NUMBERED_LIST! \
       )                                        \
       BOOST_PP_ENUM_PARAMS(i, elem)            \
       BOOST_PP_COMMA_IF(i) special             \
       BOOST_PP_REPEAT(                         \
          BOOST_PP_SUB(BOOST_PP_DEC(n), i),     \
          SPECIAL_NUMBERED_LIST_HELPER,         \
          (elem, i)                             \
       )                                        \
       /**/

    #define SPECIAL_NUMBERED_LIST_HELPER(z, i, elem_base) \
       ,                                          \
       BOOST_PP_CAT(                              \
          BOOST_PP_TUPLE_ELEM(2, 0, elem_base),   \
          BOOST_PP_ADD(                           \
             i,                                   \
             BOOST_PP_TUPLE_ELEM(2, 1, elem_base) \
          )                                       \
       )                                          \
       /**/

    SPECIAL_NUMBERED_LIST(3, 0, E, S)
    SPECIAL_NUMBERED_LIST(3, 1, E, S)
    SPECIAL_NUMBERED_LIST(3, 2, E, S)
    SPECIAL_NUMBERED_LIST(3, 3, E, S)

© Copyright Housemarque Oy 2002
Permission to copy, use, modify, sell and distribute this document is granted provided this copyright notice appears in all copies. This document is provided "as is" without express or implied warranty and with no claim as to its suitability for any purpose.

© Copyright Housemarque Oy 2002 © Copyright Paul Mensonides 2002

Distributed under the Boost Software License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at www.boost.org/LICENSE_1_0.txt)